@@ -1,10 +1,4 @@
 """
-This script will dereplicate GTDB assemblies to a user-specified threshold, copying the
-dereplicated assemblies to a new directory.
-
-Usage:
-    dereplicate_assemblies.py --threshold 0.005 assemblies derep bac_and_arc_taxonomy_r86.tsv
-
 This program is free software: you can redistribute it and/or modify it under the terms of the GNU
 General Public License as published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.
@@ -17,6 +11,7 @@ You should have received a copy of the GNU General Public License along with thi
 see <https://www.gnu.org/licenses/>.
 """
 
+from scipy import stats
 from derep_genomes import __version__
 
 import subprocess
@@ -32,8 +27,9 @@ from derep_genomes.general import (
     get_assembly_filename,
     applyParallel,
     is_debug,
+    get_assembly_length,
 )
-from derep_genomes.graph import dereplicate
+from derep_genomes.graph import dereplicate_ANI, dereplicate_mash, concat_df
 import logging
 import pathlib
 import sqlite3
@@ -51,6 +47,7 @@ from derep_genomes.dbops import (
     check_if_done,
     retrieve_all_taxa_analyzed,
     retrieve_all_jobs_done,
+    retrieve_all_jobs_failed,
     delete_from_db,
     retrieve_all_genomes_derep,
 )
@@ -58,11 +55,9 @@ from multiprocessing import Pool
 import tqdm
 from functools import partial
 import pandas as pd
-import time
 import tempfile
 
 log = logging.getLogger("my_logger")
-timestr = time.strftime("%Y%m%d-%H%M%S")
 
 
 def process_one_taxon(taxon, parms):
@@ -73,6 +68,10 @@ def process_one_taxon(taxon, parms):
     slurm_config = parms["slurm_config"]
     tmp_dir = parms["tmp_dir"]
     max_jobs_array = parms["max_jobs_array"]
+    mash_threshold = parms["mash_threshold"]
+    ani_fraglen_fraction = parms["ani_fraglen_fraction"]
+    min_genome_size = parms["min_genome_size"]
+    failed_df = None
 
     log.debug("Dereplicating {}".format(taxon))
     classification = classification[classification["taxon"] == taxon]
@@ -88,73 +87,197 @@ def process_one_taxon(taxon, parms):
                 n_assemblies
             )
         )
-        derep_assemblies, results = dereplicate(
+
+        log.debug("Preclustering with MASH...")
+        acc_to_assemblies_mash = dereplicate_mash(
             all_assemblies=acc_to_assemblies,
             threads=threads,
+            tmp_dir=tmp_dir,
+            mash_threshold=mash_threshold,
+            threshold=threshold,
+        )
+        acc_to_assemblies_mash_r = acc_to_assemblies_mash[
+            acc_to_assemblies_mash["representative"] == 1
+        ].copy()
+        acc_to_assemblies_mash_d = acc_to_assemblies_mash[
+            acc_to_assemblies_mash["derep"] == 1
+        ].copy()
+        acc_to_assemblies_mash_o = acc_to_assemblies_mash[
+            acc_to_assemblies_mash["derep"] == 0
+        ].copy()
+
+        log.debug(
+            "Identified {:,} linked assemblies in {:,} MASH communities".format(
+                acc_to_assemblies_mash_d.shape[0], acc_to_assemblies_mash_r.shape[0]
+            )
+        )
+
+        # We refine each MASH community
+        log.debug(
+            "Refining {:,} MASH selected assemblies".format(
+                acc_to_assemblies_mash_d.shape[0]
+            )
+        )
+        derep_assemblies, results, failed, stats_df = dereplicate_ANI(
+            all_assemblies=acc_to_assemblies_mash_d,
+            threads=threads,
             threshold=threshold,
             chunks=chunks,
             slurm_config=slurm_config,
             tmp_dir=tmp_dir,
             max_jobs_array=max_jobs_array,
+            min_genome_size=min_genome_size,
+            ani_fraglen_fraction=ani_fraglen_fraction,
+        )
+        if failed:
+            failed_df = acc_to_assemblies.copy()
+            failed_df.loc[:, "file"] = failed_df["assembly"].apply(os.path.basename)
+            failed_df.loc[:, "taxon"] = taxon
+            failed_df.loc[:, "reason"] = failed
+            failed_df = failed_df[["taxon", "accession", "file", "reason"]]
+        else:
+            # Add taxa to taxa table
+            derep_assemblies = concat_df([derep_assemblies, acc_to_assemblies_mash_o])
+            derep_assemblies.loc[:, "taxon"] = taxon
+            derep_assemblies.loc[:, "file"] = derep_assemblies["assembly"].apply(
+                os.path.basename
+            )
+            results.loc[:, "taxon"] = taxon
+            if stats_df is not None:
+                stats_df.loc[:, "taxon"] = taxon
+
+        return derep_assemblies, results, failed_df, stats_df
+
+
+def insert_to_db(
+    derep_assemblies, results, failed, stats_df, con, threads, out_dir, copy
+):
+    if not derep_assemblies.empty:
+        query = check_existing(
+            df=derep_assemblies.loc[:, ["taxon"]].drop_duplicates(),
+            columns=["taxon"],
+            table="taxa",
+            con=con,
+        )
+        insert_pd_sql(query=query, table="taxa", con=con)
+
+        # Add genomes to genome table
+        query = check_existing(
+            df=derep_assemblies.loc[:, ["taxon", "accession"]],
+            columns=["taxon", "accession"],
+            table="genomes",
+            con=con,
+        )
+        genomes = insert_pd_sql(query=query, table="genomes", con=con)
+
+        # Add derep genomes to derep genome table
+        query = check_existing(
+            df=derep_assemblies[derep_assemblies["derep"] == 1]
+            .loc[:, ["accession", "representative"]]
+            .copy(),
+            columns=["accession", "representative"],
+            table="genomes_derep",
+            con=con,
+        )
+        genomes_derep = insert_pd_sql(query=query, table="genomes_derep", con=con)
+
+    if not results.empty:
+        # Add results to results table
+        query = check_existing(
+            results.loc[
+                :, ["taxon", "weight", "communities", "n_genomes", "n_genomes_derep"]
+            ],
+            columns=["taxon", "weight", "communities", "n_genomes", "n_genomes_derep"],
+            table="results",
+            con=con,
+        )
+        results_db = insert_pd_sql(query=query, table="results", con=con)
+
+    if not stats_df.empty:
+        # Add results to results table
+        query = check_existing(
+            stats_df.loc[
+                :,
+                [
+                    "taxon",
+                    "representative",
+                    "n_nodes",
+                    "n_nodes_selected",
+                    "n_nodes_discarded",
+                    "graph_avg_weight",
+                    "graph_sd_weight",
+                    "graph_avg_weight_raw",
+                    "graph_sd_weight_raw",
+                    "subgraph_selected_avg_weight",
+                    "subgraph_selected_sd_weight",
+                    "subgraph_selected_avg_weight_raw",
+                    "subgraph_selected_sd_weight_raw",
+                    "subgraph_discarded_avg_weight",
+                    "subgraph_discarded_sd_weight",
+                    "subgraph_discarded_avg_weight_raw",
+                    "subgraph_discarded_sd_weight_raw",
+                ],
+            ],
+            columns=[
+                "taxon",
+                "representative",
+                "n_nodes",
+                "n_nodes_selected",
+                "n_nodes_discarded",
+                "graph_avg_weight",
+                "graph_sd_weight",
+                "graph_avg_weight_raw",
+                "graph_sd_weight_raw",
+                "subgraph_selected_avg_weight",
+                "subgraph_selected_sd_weight",
+                "subgraph_selected_avg_weight_raw",
+                "subgraph_selected_sd_weight_raw",
+                "subgraph_discarded_avg_weight",
+                "subgraph_discarded_sd_weight",
+                "subgraph_discarded_avg_weight_raw",
+                "subgraph_discarded_sd_weight_raw",
+            ],
+            table="stats",
+            con=con,
         )
 
-        # Add taxa to taxa table
-        derep_assemblies.loc[:, "taxon"] = taxon
-        derep_assemblies.loc[:, "file"] = derep_assemblies["assembly"].apply(
-            os.path.basename
+        stats_db = insert_pd_sql(query=query, table="stats", con=con)
+
+    if not derep_assemblies.empty and not results.empty:
+        # Add jobs done to table
+        query = check_existing(
+            derep_assemblies.loc[:, ["taxon", "accession", "file"]],
+            columns=["taxon", "accession", "file"],
+            table="jobs_done",
+            con=con,
         )
-        results.loc[:, "taxon"] = taxon
-        return derep_assemblies, results
+        jobs_done_db = insert_pd_sql(query=query, table="jobs_done", con=con)
 
+    # Add failed jobs to table
+    if not failed.empty:
+        query = check_existing(
+            failed.loc[:, ["taxon", "accession", "file", "reason"]],
+            columns=["taxon", "accession", "file", "reason"],
+            table="jobs_failed",
+            con=con,
+        )
+        jobs_failed = insert_pd_sql(query=query, table="jobs_failed", con=con)
 
-def insert_to_db(derep_assemblies, results, con, threads, out_dir, copy):
-    query = check_existing(
-        df=derep_assemblies.loc[:, ["taxon"]].drop_duplicates(),
-        columns=["taxon"],
-        table="taxa",
-        con=con,
-    )
-    insert_pd_sql(query=query, table="taxa", con=con)
+        query = check_existing(
+            df=failed.loc[:, ["taxon"]].drop_duplicates(),
+            columns=["taxon"],
+            table="taxa",
+            con=con,
+        )
+        insert_pd_sql(query=query, table="taxa", con=con)
 
-    # Add genomes to genome table
-    query = check_existing(
-        df=derep_assemblies.loc[:, ["taxon", "accession"]],
-        columns=["taxon", "accession"],
-        table="genomes",
-        con=con,
-    )
-    genomes = insert_pd_sql(query=query, table="genomes", con=con)
-
-    # Add derep genomes to derep genome table
-    query = check_existing(
-        df=derep_assemblies[derep_assemblies["derep"] == 1]
-        .loc[:, ["accession", "representative"]]
-        .copy(),
-        columns=["accession", "representative"],
-        table="genomes_derep",
-        con=con,
-    )
-    genomes_derep = insert_pd_sql(query=query, table="genomes_derep", con=con)
-
-    # Add results to results table
-    query = check_existing(
-        results.loc[
-            :, ["taxon", "weight", "communities", "n_genomes", "n_genomes_derep"]
-        ],
-        columns=["taxon", "weight", "communities", "n_genomes", "n_genomes_derep"],
-        table="results",
-        con=con,
-    )
-    results_db = insert_pd_sql(query=query, table="results", con=con)
-
-    # Add jobs done to table
-    query = check_existing(
-        derep_assemblies.loc[:, ["taxon", "accession", "file"]],
-        columns=["taxon", "accession", "file"],
-        table="jobs_done",
-        con=con,
-    )
-    jobs_done_db = insert_pd_sql(query=query, table="jobs_done", con=con)
+        query = check_existing(
+            failed.loc[:, ["taxon", "accession", "file"]],
+            columns=["taxon", "accession", "file"],
+            table="jobs_done",
+            con=con,
+        )
+        jobs_failed = insert_pd_sql(query=query, table="jobs_done", con=con)
 
     try:
         con.commit()
@@ -306,10 +429,13 @@ def find_assemblies(x, classifications, all_assm):
     return res
 
 
+# TODO: Improve the matching between filenames and accessions
 def shorten_accession(accession):
     if accession.startswith("GCF_") or accession.startswith("GCA_"):
         accession = accession.split(".")[0]
         assert len(accession) == 13
+    else:
+        accession = accession.split(".")[0]
     return accession
 
 
@@ -354,6 +480,7 @@ def main():
     logging.getLogger("my_logger").setLevel(
         logging.DEBUG if args.debug else logging.INFO
     )
+    prefix = args.prefix
 
     if args.copy:
         out_dir = pathlib.Path(args.out_dir).absolute()
@@ -394,6 +521,7 @@ def main():
     )
     all_assemblies_df = list(map(get_accession, all_assemblies))
     assm_data = classifications_df.merge(pd.DataFrame(all_assemblies_df))
+
     assm_data.loc[:, "file"] = assm_data["assembly"].apply(os.path.basename)
 
     if args.selected_taxa:
@@ -446,7 +574,7 @@ def main():
                         )
                     )
 
-            rfname = timestr + "-derep-genomes_removed-db-entries.tsv"
+            rfname = prefix + "-derep-genomes_removed-db-entries.tsv"
             removed.to_csv(rfname, index=False, sep="\t")
 
     else:
@@ -457,6 +585,20 @@ def main():
         assm_min = 2
         assm_max = 10
 
+        # log.info("Calculating assembly lengths")
+
+        # p = Pool(args.threads)
+        # to_do.loc[:, "len"] = list(
+        #     tqdm.tqdm(
+        #         p.imap(get_assembly_length, to_do["assembly"].to_list()),
+        #         total=to_do.shape[0],
+        #         leave=False,
+        #         ncols=80,
+        #     )
+        # )
+        # print(to_do)
+        # exit(0)
+
         log.info(
             "Splitting taxa in groups of singleton, between {}-{} and more than {} assemblies\n".format(
                 assm_min, assm_max, assm_max
@@ -489,7 +631,6 @@ def main():
         ].reset_index(drop=True)
 
         if not classification_singletons.empty:
-
             log.info(
                 "Processing {:,} singletons".format(classification_singletons.shape[0])
             )
@@ -514,6 +655,9 @@ def main():
                 "slurm_config": None,
                 "tmp_dir": tmp_dir,
                 "max_jobs_array": args.max_jobs_array,
+                "mash_threshold": args.mash_threshold,
+                "min_genome_size": args.min_genome_size,
+                "ani_fraglen_fraction": args.ani_fraglen_fraction,
             }
 
             if args.threads > 2:
@@ -527,7 +671,7 @@ def main():
             )
 
             if is_debug():
-                files = list(map(partial(process_one_taxon, parms=parms), taxons))
+                dfs = list(map(partial(process_one_taxon, parms=parms), taxons))
             else:
                 p = Pool(nworkers)
                 dfs = list(
@@ -541,11 +685,38 @@ def main():
                     )
                 )
 
-            derep_assemblies = pd.concat([x[0] for x in dfs])
-            results = pd.concat([x[1] for x in dfs])
+            l = [x[2] for x in dfs if None not in dfs]
+            if all(v is None for v in l):
+                failed = pd.DataFrame()
+            else:
+                failed = pd.concat([x[2] for x in dfs if not None in dfs])
+
+            l = [x[0] for x in dfs if None not in dfs]
+
+            if all(v is None for v in l):
+                derep_assemblies = pd.DataFrame()
+            else:
+                derep_assemblies = pd.concat([x[0] for x in dfs if not None in dfs])
+
+            l = [x[1] for x in dfs if None not in dfs]
+
+            if all(v is None for v in l):
+                results = pd.DataFrame()
+            else:
+                results = pd.concat([x[1] for x in dfs if not None in dfs])
+
+            l = [x[3] for x in dfs if None not in dfs]
+
+            if all(v is None for v in l):
+                stats_df = pd.DataFrame()
+            else:
+                stats_df = pd.concat([x[3] for x in dfs if not None in dfs])
+
             insert_to_db(
                 derep_assemblies=derep_assemblies,
                 results=results,
+                failed=failed,
+                stats_df=stats_df,
                 con=con,
                 threads=args.threads,
                 out_dir=args.out_dir,
@@ -575,6 +746,9 @@ def main():
                 "slurm_config": args.slurm_config.name,
                 "tmp_dir": tmp_dir,
                 "max_jobs_array": args.max_jobs_array,
+                "mash_threshold": args.mash_threshold,
+                "min_genome_size": args.min_genome_size,
+                "ani_fraglen_fraction": args.ani_fraglen_fraction,
             }
 
             if is_debug():
@@ -590,35 +764,66 @@ def main():
                         ncols=80,
                     )
                 )
-            derep_assemblies = pd.concat([x[0] for x in dfs])
-            results = pd.concat([x[1] for x in dfs])
+
+            l = [x[2] for x in dfs if None not in dfs]
+            if all(v is None for v in l):
+                failed = pd.DataFrame()
+            else:
+                failed = pd.concat([x[2] for x in dfs if not None in dfs])
+
+            l = [x[0] for x in dfs if None not in dfs]
+            if all(v is None for v in l):
+                derep_assemblies = pd.DataFrame()
+            else:
+                derep_assemblies = pd.concat([x[0] for x in dfs if not None in dfs])
+
+            l = [x[1] for x in dfs if None not in dfs]
+            if all(v is None for v in l):
+                results = pd.DataFrame()
+            else:
+                results = pd.concat([x[1] for x in dfs if not None in dfs])
+            if all(v is None for v in l):
+                stats_df = pd.DataFrame()
+            else:
+                stats_df = pd.concat([x[3] for x in dfs if not None in dfs])
             insert_to_db(
                 derep_assemblies=derep_assemblies,
                 results=results,
+                failed=failed,
+                stats_df=stats_df,
                 con=con,
                 threads=args.threads,
                 out_dir=args.out_dir,
                 copy=args.copy,
             )
-
         jobs_done = retrieve_all_jobs_done(con)
+        jobs_failed = retrieve_all_jobs_failed(con)
         genomes_derep = retrieve_all_genomes_derep(con)
-        jobs_done = jobs_done.merge(genomes_derep)
-        fname = timestr + "-derep-genomes_results.tsv"
-        logging.info("Saving results to {}".format(fname))
-        jobs_done.loc[:, "src"] = jobs_done["file"].apply(
-            lambda x: os.path.join(in_dir, x)
-        )
-        if not args.copy:
-            jobs_done.loc[:, "dst"] = ""
-        else:
-            jobs_done.loc[:, "dst"] = jobs_done["file"].apply(
-                lambda x: os.path.join(out_dir, x)
+        logging.info(
+            "Jobs failed: {} / Jobs done: {}".format(
+                jobs_failed.shape[0], jobs_done.shape[0]
             )
-        jobs_done["representative"] = jobs_done["representative"].astype(bool)
-        jobs_done[["taxon", "accession", "representative", "src", "dst"]].to_csv(
-            fname, index=False, sep="\t"
         )
+        if not jobs_done.empty and not genomes_derep.empty:
+            jobs_done = retrieve_all_jobs_done(con)
+            genomes_derep = retrieve_all_genomes_derep(con)
+
+            jobs_done = jobs_done.merge(genomes_derep)
+            fname = prefix + "-derep-genomes_results.tsv"
+            logging.info("Saving results to {}".format(fname))
+            jobs_done.loc[:, "src"] = jobs_done["file"].apply(
+                lambda x: os.path.join(in_dir, x)
+            )
+            if not args.copy:
+                jobs_done.loc[:, "dst"] = ""
+            else:
+                jobs_done.loc[:, "dst"] = jobs_done["file"].apply(
+                    lambda x: os.path.join(out_dir, x)
+                )
+            jobs_done["representative"] = jobs_done["representative"].astype(bool)
+            jobs_done[["taxon", "accession", "representative", "src", "dst"]].to_csv(
+                fname, index=False, sep="\t"
+            )
     else:
         logging.info("Didn't find any new assemblies to process")
     con.close()
